@@ -3,7 +3,7 @@ import numpy as np
 from requests import post
 from dataclasses import dataclass, fields
 from operator import attrgetter
-from typing import Optional, List, Dict
+from typing import Optional, List, Dict, Callable
 from pickle import loads
 from gzip import decompress
 from base64 import b64decode
@@ -70,6 +70,44 @@ class EmbeddingCollection:
             )
         )
 
+@dataclass
+class CoefPredicate:
+
+    predicate: Callable[[KeyValue], bool]
+    coefficient: float
+
+    def __call__(self, embedding_component: Embedding) -> Embedding:
+        if self.predicate(embedding_component.key_value):
+            return Embedding(
+                vector=embedding_component.vector*self.coefficient,
+                key_value=embedding_component.key_value,
+            )
+        return embedding_component
+    
+@dataclass
+class PredicateGroup:
+
+    name: str
+    predicate: Callable[[KeyValue], bool]
+
+@dataclass
+class GroupEmbedding:
+
+    name: str
+    embeddings: List[Embedding]
+
+    def to_array(self, agg = lambda x: x) -> np.ndarray:
+        return agg(
+            np.array(
+                list(
+                    map(
+                        lambda x: x.vector,
+                        self.embeddings
+                    )
+                )
+            )
+        )
+
 @dataclass
 class EmbeddingCollectionComposer:
 
@@ -112,6 +150,69 @@ class EmbeddingCollectionComposer:
                 )
             )
         )
+    
+    def scale(self, predicates: List[CoefPredicate]) -> "EmbeddingCollectionComposer":
+        """
+            Scales embeddings relative with respect to the predicates.
+            If an embedding matches multiple predicates, all of the coefficients are multiplied.
+        """
+        return EmbeddingCollectionComposer(
+            groups=self.groups,
+            embeddings=dict(
+                starmap(
+                    lambda key, embedding: (
+                        key,
+                        Embedding(
+                            vector=embedding.vector * np.prod(
+                                list(
+                                    map(
+                                        lambda cpred: cpred.coefficient,
+                                        filter(
+                                            lambda cpred: cpred.predicate(embedding.key_value),
+                                            predicates
+                                        )
+                                    )
+                                )
+                            ),
+                            key_value=embedding.key_value,
+                        )
+                    ),
+                    self.embeddings.items(),
+                )
+            )
+        )
+    
+    def group(self, predicate_groups: List[PredicateGroup]) -> List[List[GroupEmbedding]]:
+
+        """
+            Groups embeddings by the predicate groups.
+            If no match from predicate groups, the embedding is not included in the result.
+
+            :param predicate_groups: A list of predicate groups.
+            :return: A list of lists of group embeddings.
+        """
+
+        new_groups = []
+        for group in self.groups:
+            new_sub_groups = {}
+            for embedding in group:
+                for predicate_group in predicate_groups:
+                    new_sub_groups.setdefault(predicate_group.name, [])
+                    if predicate_group.predicate(self.embeddings[embedding].key_value):
+                        new_sub_groups[predicate_group.name].append(self.embeddings[embedding])
+            new_groups.append(
+                list(
+                    map(
+                        lambda k: GroupEmbedding(
+                            name=k,
+                            embeddings=new_sub_groups[k],
+                        ),
+                        new_sub_groups,
+                    )
+                )
+            )
+
+        return new_groups                
 
 @dataclass
 class Result:
